/* Help content is generated automatically from README.adoc by helpgen script */

static const char* helpInfo =
"<qt>\n"
"<center><h1>QGit Handbook</h1></center>\n"
"<h2><a name=\"_command_line_arguments\"></a>Command line arguments</h2>\n"
"<p>Run qgit from a git working directory, command line arguments\n"
"are filtered by <em>git log</em>. Some examples:</p>\n"
"<table border=\"0\" bgcolor=\"#e8e8e8\" width=\"100%\" cellpadding=\"10\"><tr><td><!-- Generator: GNU source-highlight\n"
"by Lorenzo Bettini\n"
"http://www.lorenzobettini.it\n"
"http://www.gnu.org/software/src-highlite -->\n"
"<pre><tt>qgit --no-merges\n"
"qgit v2<font color=\"#990000\">.</font><font color=\"#993399\">6.18</font><font color=\"#990000\">..</font> include/scsi drivers/scsi\n"
"qgit --since<font color=\"#990000\">=</font><font color=\"#FF0000\">\"2 weeks ago\"</font> -- kernel<font color=\"#990000\">/</font>\n"
"qgit -r --name-status release<font color=\"#990000\">..</font><b><font color=\"#0000FF\">test</font></b></tt></pre></td></tr></table>\n"
"<p>If qgit is launched without arguments or if you change archive with\n"
"<em>open</em> menu, a dialog for range select is shown.\n"
"You can select top and bottom rev tags from the list or paste a\n"
"specific revision. Values are passed to <em>git log</em> to narrow\n"
"data loading to chosen revisions.</p>\n"
"<hr>\n"
"<h2><a name=\"_main_view\"></a>Main view</h2>\n"
"<p>You can navigate through logs, file names, file history, archive tree.\n"
"All the views will be updated accordingly.</p>\n"
"<p>Copy/paste is supported on all fields. Copy (CTRL+C) is supported on\n"
"all views.</p>\n"
"<p>All the references found recursively under .git/refs/ directory are\n"
"highlighted according to their type: current branch(HEAD), branch, tag,\n"
"other. Reference names and any associated messages can be viewed in status\n"
"bar when a tagged revision is selected.</p>\n"
"<p>When you right click on main view a context sensitive pop-up menu is shows\n"
"available commands and a <em>quick jump</em> tag list.</p>\n"
"<dl>\n"
"<dt>\n"
"<strong>Key bindings</strong>\n"
"</dt>\n"
"<dd>\n"
"</dd>\n"
"</dl>\n"
"<div>\n"
"<table rules=\"all\"\n"
"width=\"100%\"\n"
"frame=\"border\"\n"
"cellspacing=\"0\" cellpadding=\"4\">\n"
"<tbody>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>r</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Go to revisions list page</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>p</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Go to patch page</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>f</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Go to file page</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;Alt+wheel&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Go to next/previous page</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>t</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Toggle tree view</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>s</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Toggle view of secondary panes</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>h</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Toggle view of revision header</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;Home&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Move to first revision</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;End&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Move to last revision</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>i</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Move up one revision in main view (global scope)</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>n, k</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Move down one revision in main view (global scope)</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;Shift-Up&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Move to previous highlighted line</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;Shift-Down&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Move to next highlighted line</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;Left&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Go back in history list</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;Right&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Go forward in history list</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;CTRL-plus&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Increase font size</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;CTRL-minus&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Decrease font size</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;Delete&gt;, b</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Scroll content up one page</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;Backspace&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Scroll content up one page</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>&lt;Space&gt;</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Scroll content down one page</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>u</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Scroll content up 18 lines</p></td>\n"
"</tr>\n"
"<tr>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>d</p></td>\n"
"<td align=\"left\" width=\"50%\" valign=\"top\"><p>Scroll content down 18 lines</p></td>\n"
"</tr>\n"
"</tbody>\n"
"</table>\n"
"</div>\n"
"<dl>\n"
"<dt>\n"
"<strong>Directory tree</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"From menu or toolbar button it is possible to show a side panel with\n"
"tree view of repository files and directories.\n"
"</p>\n"
"<p>Double clicking on a file opens file annotation window. With <em>filter by tree</em>\n"
"button it is possible to compress revision list to show only selected\n"
"files/directories in tree view.</p>\n"
"<p>Tree view supports multi-selection. When you right click on a file on tree\n"
"view a context sensitive pop-up menu is shows with available commands.</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Working directory changes</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"When <em>Check working dir</em> flag is set, as example from main view context\n"
"pop-up menu, a pseudo-revision is shown and highlighted at the top of the\n"
"list. Highlight and revision name reflect current working directory status:\n"
"<em>Nothing to commit</em> or <em>Working directory changes</em> respectively.\n"
"</p>\n"
"<p>To check for working directory modified files set corresponding preference in\n"
"Edit&#8594;Settings&#8594;'Working dir'. QGit checks for possible new files added in\n"
"working directory using ignoring rules according to git ls-files\n"
"specifications, see menu Edit&#8594;Settings&#8594;'Working dir'.</p>\n"
"<table frame=\"void\" cellpadding=\"4\">\n"
"<tr valign=\"top\">\n"
"<td>\n"
"<p><b><u>Tip</u></b></p>\n"
"</td>\n"
"<td style=\"border-left: 1px solid silver;\">If you don&#8217;t need to see modified files in working dir, disable\n"
"corresponding setting and start-up time will be shorter.</td></tr></table>\n"
"</dd>\n"
"<dt>\n"
"<strong>Lane info</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"Selecting a lane with mouse right button will display a pop-up\n"
"with the list of children and parent. Select one and you jump to it.\n"
"</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Filter / Highlight</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"Use the combo box to select where you want to filter or highlight on.\n"
"Currently supported fields are: log header, log message, revision author,\n"
"revision SHA1, file name and patch content.\n"
"</p>\n"
"<p>Write a filter string, press filter button and the view\n"
"will update showing only commits that contain the filter string,\n"
"case insensitive. Toggle filter button to release the filter.</p>\n"
"<p>Alternatively press the magnifying glass button, in this case matched\n"
"lines will be highlighted, you can use &lt;Shift-Up&gt; and &lt;Shift-Down&gt;\n"
"keys to browse them. Toggle the button to remove the highlighting.</p>\n"
"<table frame=\"void\" cellpadding=\"4\">\n"
"<tr valign=\"top\">\n"
"<td>\n"
"<p><b><u>Note</u></b></p>\n"
"</td>\n"
"<td style=\"border-left: 1px solid silver;\">In case of patch content regexp filtering, the given string is\n"
"interpreted as a POSIX regular expression, not as a simple substring.</td></tr></table>\n"
"<table frame=\"void\" cellpadding=\"4\">\n"
"<tr valign=\"top\">\n"
"<td>\n"
"<p><b><u>Tip</u></b></p>\n"
"</td>\n"
"<td style=\"border-left: 1px solid silver;\">Very useful to quick retrieve a sha writing only first 3-4\n"
"digits and filtering / highlighting on revision sha. The sha value\n"
"can then be copied from SHA field.</td></tr></table>\n"
"<table frame=\"void\" cellpadding=\"4\">\n"
"<tr valign=\"top\">\n"
"<td>\n"
"<p><b><u>Tip</u></b></p>\n"
"</td>\n"
"<td style=\"border-left: 1px solid silver;\">It is possible to insert an abbreviated sha directly in the\n"
"SHA line edit at the top right of the window. After pressing enter\n"
"this will trigger an higlighting of the matched revisions. It is\n"
"a kind of shortcut of the previous tip.</td></tr></table>\n"
"</dd>\n"
"<dt>\n"
"<strong>Save patch series</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"After mouse selecting the chosen revisions (use standard CTRL+left click)\n"
"for single select or SHIFT+left click for range select), press <em>Save\n"
"Patch</em> button or use file menu and a dialog will let you choose patches\n"
"destination directory. Then <em>git format-patch-script</em> will be called and\n"
"patches created. It is possible to specify additional options with\n"
"Edit&#8594;Settings menu.\n"
"</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Apply patch</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"This menu entry is complementary to save patch and it&#8217;s an interface\n"
"to <em>git am</em>.\n"
"</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Drag and drop</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"It is possible to drag some selected revs from one instance of qgit to another\n"
"open on a different archive. In this case <em>git format-patch</em> is used in the\n"
"dragging archive to create temporary patches imported in the dropping archive\n"
"by <em>git am</em>.\n"
"</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Make branch</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"Select a revision and open Edit&#8594;'Make Branch' or use right click context\n"
"pop-up menu. A dialog will be shown asking for a branch name.\n"
"</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Make tag</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"Select a revision and open Edit&#8594;'Make Tag' or use right click context\n"
"pop-up menu. Two dialogs will be shown, the first asking for a tag name, the\n"
"second for a tag message (not mandatory). If a non empty message is written,\n"
"this will be saved together with the tag. Tags and tag messages can be viewed\n"
"in status bar when a tagged revision is selected.\n"
"</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Delete tag</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"Select a tagged revision and open Edit&#8594;'Delete Tag' or use right click\n"
"context pop-up menu. After confirmation the selected revision will be\n"
"untagged.\n"
"</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Save file</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"Select a file from tree or file list and open File&#8594;'Save file as' or use the\n"
"tree view context sensitive pop-up menu (right click), a dialog will be shown\n"
"asking for a file name (default to current) and destination directory. Input\n"
"a valid name, press OK and the file will be saved.\n"
"</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Commit changes</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"When enabled with Edit&#8594;Settings&#8594;'Working dir'&#8594;'Diff against working dir'\n"
"and there is something committable, a special highlighted first revision is\n"
"shown, with the status of the archive and the possible pending stuff.\n"
"From Edit&#8594;Commit it is then possible to invoke the commit dialog.\n"
"</p>\n"
"<p>In commit dialog select the files to commit or, simply, to sync with index\n"
"(call <em>git update-index</em> on them). A proper commit message may be entered and,\n"
"after confirmation, changes are committed and a new revision is created.</p>\n"
"<p>It is also possible to amend last commit. The Edit&#8594;Amend commit opens the\n"
"same dialog, but changes are added to the head commit instead of creating new\n"
"commit.</p>\n"
"<p>The core commit function is performed by <em>git commit</em>.</p>\n"
"<table frame=\"void\" cellpadding=\"4\">\n"
"<tr valign=\"top\">\n"
"<td>\n"
"<p><b><u>Tip</u></b></p>\n"
"</td>\n"
"<td style=\"border-left: 1px solid silver;\">It is possible to use a template for commit message, use\n"
"Edit&#8594;Settings&#8594;Commit to define template file path.</td></tr></table>\n"
"</dd>\n"
"</dl>\n"
"<hr>\n"
"<h2><a name=\"_patch_viewer\"></a>Patch viewer</h2>\n"
"<p>To open patch tab use context menu, double click on a revision or file in\n"
"main view or select View&#8594;'View patch' menu (CTRL+P). The patch shown is\n"
"the diff of current selected commit against:</p>\n"
"<ul>\n"
"<li>\n"
"<p>\n"
"Parent (default)\n"
"</p>\n"
"</li>\n"
"<li>\n"
"<p>\n"
"HEAD\n"
"</p>\n"
"</li>\n"
"<li>\n"
"<p>\n"
"Selected SHA or reference name\n"
"</p>\n"
"</li>\n"
"</ul>\n"
"<p>In the last case SHA is chosen by writing or pasting a tree-ish or a reference\n"
"names in the corresponding field and pressing return. You get the same result\n"
"also with a CTRL+right click on a revision in main list. Selected target\n"
"will be highlighted. CTRL+right click again on the highlighted revision to\n"
"release the filter.</p>\n"
"<p>With the <em>filter</em> button at the right of the tool bar it is possible to\n"
"toggle the display of removed code lines. This can be useful to easy\n"
"reading of the patch.</p>\n"
"<hr>\n"
"<h2><a name=\"_external_diff_tool\"></a>External diff tool</h2>\n"
"<p>From <em>View&#8594;External diff</em> it is possible to invoke an external diff tool,\n"
"as example to view the diffs in a two vertical tiled windows.</p>\n"
"<p>External diff tool shows the diffs between two files.\n"
"First file is the current selected file of current revision.\n"
"Second file is the same file of the parent revision or of a specific revision\n"
"if <em>diff to sha</em> feature is enabled (diff target is highlighted, see above).</p>\n"
"<p>Default external viewer is kompare, but it is possible to set a preferred one\n"
"from <em>Edit&#8594;Settings&#8594;External Diff Tool</em>.</p>\n"
"<hr>\n"
"<h2><a name=\"_file_viewer\"></a>File viewer</h2>\n"
"<p>It is possible to view file contents of any file at any revision time in\n"
"history.</p>\n"
"<dl>\n"
"<dt>\n"
"<strong>File list panel</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"In the bottom right of main view a list of files modified by current\n"
"revision is shown. Selecting a file name will update the patch view\n"
"to center on the file. File names colors use the following convention\n"
"</p>\n"
"<ul>\n"
"<li>\n"
"<p>\n"
"black for modified files\n"
"</p>\n"
"</li>\n"
"<li>\n"
"<p>\n"
"green for new files\n"
"</p>\n"
"</li>\n"
"<li>\n"
"<p>\n"
"red for removed files\n"
"</p>\n"
"</li>\n"
"<li>\n"
"<p>\n"
"dark blue for renamed/copied files\n"
"</p>\n"
"</li>\n"
"</ul>\n"
"</dd>\n"
"<dt>\n"
"<strong>Merge files</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"In case of merges the groups of files corresponding to each merge parent\n"
"are separated by two empty lines.\n"
"</p>\n"
"<p>In case of merges you can chose between to see all the merge files or only\n"
"the interesting ones (default), i.e. the files modified by more then one\n"
"merge parent.</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>File content</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"To view file content double click on a file name in tree view, or use context\n"
"menu in file list or select View&#8594;'View file' menu (CTRL+A).\n"
"</p>\n"
"<p>In file view page will be shown current revision&#8217;s file content and file\n"
"history.</p>\n"
"<p>It is possible to copy to the clipboard the selected content with CTRL+C or\n"
"with the corresponding button.</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>File annotations</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"On opening or updating file viewer, file history will be retrieved from archive\n"
"together with file content. Annotations are then calculated in background\n"
"and the view is updated when ready.\n"
"</p>\n"
"<p>Double clicking on an annotation index number will update history list\n"
"to point to corresponding revision.</p>\n"
"<p>Hovering the mouse over an annotation index will show a tool tip with the\n"
"corresponding revision description.</p>\n"
"<p>File content will change too, to show new selected revision file. To keep\n"
"the same view content after double clicking, probably what you want, just pin\n"
"it with <em>Pin view</em> check button. Next to the check button there is a spinbox\n"
"to show/select the current revision number.</p>\n"
"<p>Double click on history list entry to update main, patch and tree views to\n"
"corresponding revision.</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Code region filter</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"When annotation info is available the <em>filter</em> button is enabled and it is\n"
"possible to mouse select a region of file content. Then, when pressing\n"
"the filter button, only revisions that modify the selected region will be\n"
"visible. Selected code region is highlighted and a shrunken history is\n"
"shown. Filter button is a toggle button, so just press it again to\n"
"release the filter.\n"
"</p>\n"
"</dd>\n"
"<dt>\n"
"<strong>Syntax highlighter</strong>\n"
"</dt>\n"
"<dd>\n"
"<p>\n"
"If GNU Source-highlight (<a href=\"http://www.gnu.org/software/src-highlite/\">http://www.gnu.org/software/src-highlite/</a>) is\n"
"installed and in PATH then it is possible to toggle source code highlight\n"
"pressing the <em>Color text</em> tool button. Please refer to Source-highlight\n"
"site for the list of supported languages and additional documentation.\n"
"</p>\n"
"</dd>\n"
"</dl>\n"
"<hr>\n"
"<h2><a name=\"_actions\"></a>Actions</h2>\n"
"<p>Actions can be added/removed using a dedicated dialog invoked\n"
"from <em>Actions&#8594;Setup actions&#8230;</em> menu. Actions can be activated\n"
"clicking on their name from the Actions menu.</p>\n"
"<p>Each action can be associated to a list of any type of git or shell\n"
"commands or to an external script.</p>\n"
"<p>While an action is running a terminal window is shown to display the\n"
"corresponding output.</p>\n"
"<p>An action can also ask for command line arguments before to run so\n"
"to allow for maximum flexibility.</p>\n"
"<table frame=\"void\" cellpadding=\"4\">\n"
"<tr valign=\"top\">\n"
"<td>\n"
"<p><b><u>Note</u></b></p>\n"
"</td>\n"
"<td style=\"border-left: 1px solid silver;\">command line arguments are always appended to the first command only.\n"
"This lets you define an action like:</td></tr></table>\n"
"<table border=\"0\" bgcolor=\"#e8e8e8\" width=\"100%\" cellpadding=\"10\"><tr><td><!-- Generator: GNU source-highlight\n"
"by Lorenzo Bettini\n"
"http://www.lorenzobettini.it\n"
"http://www.gnu.org/software/src-highlite -->\n"
"<pre><tt>git fetch\n"
"git merge</tt></pre></td></tr></table>\n"
"<p>And if you type <em>origin</em> when prompted, the action executed will be:</p>\n"
"<table border=\"0\" bgcolor=\"#e8e8e8\" width=\"100%\" cellpadding=\"10\"><tr><td><!-- Generator: GNU source-highlight\n"
"by Lorenzo Bettini\n"
"http://www.lorenzobettini.it\n"
"http://www.gnu.org/software/src-highlite -->\n"
"<pre><tt>git fetch origin\n"
"git merge</tt></pre></td></tr></table>\n"
"<p>If you need a more complex arguments passing with a shell like notation\n"
"define a script and associate your action to it.</p>\n"
"<hr>\n"
"<h2><a name=\"_integration_with_stgit\"></a>Integration with StGIT</h2>\n"
"<p>When a StGIT stack is found on top of a git archive, qgit transparently\n"
"handles the added information.</p>\n"
"<p>Integration with StGIT is implemented both by new and modified functions.</p>\n"
"<p><b>New functions are automatically activated:</b></p><ul>\n"
"<li>\n"
"<p>\n"
"Visualization of applied and unapplied patches in main view.\n"
"</p>\n"
"</li>\n"
"<li>\n"
"<p>\n"
"Interface to push/pop patches by a mouse right click on selected items.\n"
"   Push supports also multi-selection.\n"
"</p>\n"
"</li>\n"
"</ul>\n"
"<p><b>Existing functions change behavior:</b></p><ul>\n"
"<li>\n"
"<p>\n"
"Amend commit dialog refreshes top stack patch with modified files instead\n"
"   of amending the commit. It is appropriately renamed in the menu.\n"
"</p>\n"
"</li>\n"
"<li>\n"
"<p>\n"
"Commit dialog creates a new patch on the top of the stack filled with\n"
"   modified working directory content instead of commit a new revision to\n"
"   git repository.\n"
"</p>\n"
"</li>\n"
"<li>\n"
"<p>\n"
"Apply patch changes to interface StGIT import and fold commands instead\n"
"   of applying patch directly on the git repository.\n"
"</p>\n"
"</li>\n"
"</ul>\n"
"</qt>\n";
