/* config.h.  Generated by configure.  */
/* config.h.in.  Generated from configure.ac by autoheader.  */

/* Name of package */
#define PACKAGE "qgit"

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT ""

/* Define to the full name of this package. */
#define PACKAGE_NAME "qgit"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "qgit"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "qgit"

/* Define to the version of this package. */
#define PACKAGE_VERSION "2.13"

/* Version number of package */
#define VERSION "2.13"
